/*!
 * @file MainWindow.cc
 * @brief 祦ɥ()
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2005-02-20, 
 * @date 2005-05-09, åǤưǤ褦ѹ
 *
 * Copyright (C) 2005, Yashiro Takeshi
 */

#include <basic.h>
#include <btron/dp.h>
#include <btron/hmi.h>
#include <btron/libapp.h>
#include <btron/file.h>
#include <keycode.h>
#include <tcode.h>
#include <bstdio.h>
#include "tcsliteral.h"
#include "MainWindow.h"
#include "Exception.h"
#include "databox.h"

extern DEV_SPEC SCREEN;

MainWindow::MainWindow()
  : bsfile(0), wid(-1), title((TC[]){TCS_WINDOWTITLE}), state(STATE_NORMAL)
{
  bgpat.spat.kind = 0;
  bgpat.spat.hsize = bgpat.spat.vsize = 1;
  bgpat.spat.fgcol = 0x90000000;
  bgpat.spat.bgcol = 0x90ffffff;
  bgpat.spat.mask = FILL50;
  setrect(wrect, 0, 0, SCREEN.hpixels, SCREEN.vpixels);
  setrect(cliprect, 0, 0, SCREEN.hpixels - scrlbar_width, SCREEN.vpixels);
}

// ɽڤؤ(ºݤˤϥɥΥץ󡦥)
void MainWindow::display(bool state = true)
{
  if (state){
    if (wid < 0){
      wid = wopn_wnd(WA_SFULL, 0, &wrect, NULL, 2, title, &bgpat, NULL);
      wreq_dsp(wid);
    }
  }
  else {
    if (wid >= 0){
      wcls_wnd(wid, CLR);
      wid = -1;
    }
  }
}

void MainWindow::draw()
{
  GID gid = wget_gid(wid);
  BMP* screen_bmp = const_cast<BMP*>(mdenv.getBMP());
  gset_vis(gid, wrect);

  static PAT barpat = {{ 0, 1, 1, 0x10585858, 0x10a0a0a0, FILL50 }};

  RECT bbound = {{ screen_bmp->bounds.c.left, screen_bmp->bounds.c.top,
                   screen_bmp->bounds.c.right - scrlbar_width,
                   screen_bmp->bounds.c.bottom }};
  RECT barrect = {{ SCREEN.hpixels - scrlbar_width, screen_bmp->bounds.c.top,
                    SCREEN.hpixels, screen_bmp->bounds.c.bottom }};

  grst_bmp(gid, &cliprect, screen_bmp, &bbound, NULL, G_STORE);
  gfil_rec(gid, barrect, &barpat, 0, G_STORE);
}

// ɽ
void MainWindow::redisp(WEVENT* ev)
{
  if (wid < 0){
    throw PaneledFailureException(EPNL_WID);
  }
  do {
    if (wsta_dsp(wid, NULL, NULL) == 0)
      break;
    //wera_wnd(wid, NULL);
    draw();
  } while (wend_dsp(wid) > 0);
  return;
}

// 
void MainWindow::onMouseButtonDown(WEVENT* ev)
{
  switch (ev->s.cmd) {
  case W_PICT:
    switch (wchk_dck(ev->s.time)) {
    case W_DCLICK:
      terminate();
    default:
      return;
    }
  case W_WORK:
    {
      DrawingStroke stroke;
      GID dgid, wgid, mgid;
      PNT ep;

      dgid = wsta_drg(wid, 0);
      wgid = wget_gid(wid);
      mgid = mdenv.getGID();
      gset_vis(dgid, cliprect);
      gset_vis(wgid, cliprect);
      stroke.addPoint(ev->s.pos);
      if (dgid < 0 || wgid < 0){
        throw PaneledFailureException(EPNL_GID);
      }
      while (wget_drg(&ep, ev) != EV_BUTUP){
        stroke.addPoint(ep);
        stroke.drawLatest(dgid);
      }
      wend_drg();

      gset_vis(mgid, cliprect);
      stroke.draw(mgid, wrect.c.top, wrect.c.bottom); // need conv
      // redisp
      doc.addStroke(stroke);
    }
    return;
  default:
    return;
  }
  return;
}

void MainWindow::onRightMouseButtonDown(WEVENT *ev)
{
  std::auto_ptr<SelectionStroke> stroke(new SelectionStroke);
  GID dgid, wgid;
  PNT ep;

  dgid = wsta_drg(wid, 0);
  wgid = wget_gid(wid);
  gset_vis(dgid, cliprect);
  gset_vis(wgid, cliprect);
  gcnv_rel(wgid, &(ev->s.pos));
  stroke->addPoint(ev->s.pos);
  if (dgid < 0 || wgid < 0){
    throw PaneledFailureException(EPNL_GID);
  }
  while ((ev->s.stat & ES_BUT2) != 0){
    wget_drg(&ep, ev);
    stroke->addPoint(ep);
    stroke->drawLatest(dgid);
  }
  stroke->finishSelection();
  wend_drg();

  std::auto_ptr<ScribbleDoc::DrawingStrokePtrList>
    selected(new ScribbleDoc::DrawingStrokePtrList);
  doc.select(selected.get(), *stroke, cliprect.c.top, cliprect.c.bottom);
  if (selected->size() <= 0){
    state = STATE_NORMAL;
    draw();
    return;
  }

  // Ȥꤢ򤵤줿Τƽľ
  mdenv.clear();
  doc.draw(mdenv.getGID(), cliprect.c.top, cliprect.c.bottom);
  draw();

  // ǡ򤵤줿Τwgid
  ScribbleDoc::DrawingStrokeConstPtrIterator iend = selected->end();
  for (ScribbleDoc::DrawingStrokePtrIterator iter = selected->begin();
       iter != iend; ++iter){
    (*iter)->drawSelection(wgid);
  }
  stroke->draw(wgid, wrect.c.top, wrect.c.bottom);

  sel_stroke = stroke;
  sel_slist = selected;
  state = STATE_SELECTED;
}

void MainWindow::onScroll(WEVENT *ev)
{
  GID dgid, wgid;
  PNT sp = ev->s.pos, tp = ev->s.pos, ep;
  int sum_dy;

  dgid = wsta_drg(wid, 0);
  wgid = wget_gid(wid);
  gset_vis(dgid, cliprect);
  gset_vis(wgid, cliprect);
  if (dgid < 0 || wgid < 0){
    throw PaneledFailureException(EPNL_GID);
  }
  while (wget_drg(&ep, ev) != EV_BUTUP){
    int dy = ep.y - tp.y;
    if (dy != 0){
      PAT pat = {{ 0, 1, 1, 0x10c0c0ff, 0x10ffffff, FILL50 }};
      gscr_rec(wgid, cliprect, 0, dy, NULL, &pat);
    }
    tp.x = ep.x;
    tp.y = ep.y;
  }
  sum_dy = ep.y - sp.y;
  wrect.c.top -= sum_dy;
  wrect.c.bottom -= sum_dy;
  cliprect.c.top -= sum_dy;
  cliprect.c.bottom -= sum_dy;
  wset_wrk(wid, &wrect);

  wend_drg();

  // Ķ˽ľ
  mdenv.setPos(cliprect.p.lefttop);
  mdenv.clear();
  doc.draw(mdenv.getGID(), cliprect.c.top, cliprect.c.bottom);

  draw();
}

// ɥν
void MainWindow::onIdle(WEVENT *ev)
{
  if (ev->s.wid != wid ||
      ev->s.cmd != W_WORK ||
      ev->s.stat & ES_CMD){
    return;
  }
  setpointer(PS_SELECT, NULL);
}

// Ͻ
void MainWindow::onKeyDown(WEVENT* ev)
{
  if (ev->s.stat & ES_CMD){
    W i = mfnd_key(menu.getMenuID(), ev->s.wid);
    if (i > 0){
      selectMenu(ev, i);
    }
    return;
  }
  // ʲɬפʤ̾省ν
  return;
}

// ˥塼ܵư
void MainWindow::selectMenu(WEVENT *ev, W mno)
{
  MainMenuItem mitem = menu.getInternalID(mno);
  
  switch (mitem){
  case MN_FINISH: // λ
    if (doc.isModified()){
      switch (panel(PNL_CLOSE)){
      case 0: // 󥻥
        break;
      case 1: // Ѵƽλ
        terminate();
        break;
      case 2: // ƽλ
        try {
          setpointer(PS_BUSY, NULL);
          doc.save(bsfile);
        }
        catch (FileOpenException& e){
          errpanel(EPNL_OPEN, 0);
          break;
        }
        catch (FileWriteException& e){
          errpanel(EPNL_WRITE, 0);
          break;
        }
        setpointer(PS_SELECT, NULL);
        terminate();
        break;
      }
    }
    else {
      terminate();
    }
    break;
  case MN_REDISP: // ɽ
    wreq_dsp(wid);
    break;
  case MN_SAVE:   // ¸
    try {
      setpointer(PS_BUSY, NULL);
      doc.save(bsfile);
    }
    catch (FileOpenException& e){
      errpanel(EPNL_OPEN, 0);
    }
    catch (FileWriteException& e){
      errpanel(EPNL_WRITE, 0);
    }
    setpointer(PS_SELECT, NULL);
    break;
  case MN_WINDOW: // ɥ
    wexe_dmn(mno);
    break;
  case MN_GADGET: // ʪ
    oexe_apg(0, mno);
    break;
  }
}

// ٥Ƚμ¹
void MainWindow::run()
{
  display(true);
  
  WEVENT wev0; 
  // ٥ȥ롼: 㳰ȤΤǼ
  for (cont = true; cont; ) {
    wget_evt(&wev0, WAIT);
    
    switch (wev0.s.type){
    case EV_NULL:
    case EV_NOMSG:
      onIdle(&wev0);
      break;
    case EV_REQUEST:
      switch (wev0.g.cmd) {
      case W_REDISP:
        redisp(&wev0);
        break;
      case W_PASTE:
        wrsp_evt(&wev0, 1); // NACK
        break;
      case W_DELETE:
      case W_FINISH:
        wrsp_evt(&wev0, 0); // ACK
        terminate();
      }
      break;
    case EV_RSWITCH:
      redisp(&wev0);
    case EV_SWITCH:
    case EV_BUTDWN:
      if (wev0.s.pos.x >= SCREEN.hpixels - scrlbar_width){
        switch (state){
        case STATE_NORMAL:
          onScroll(&wev0);
          break;
        case STATE_SELECTED:
          cancelSelection();
          break;
        }
      }
      else {
        switch (state){
        case STATE_NORMAL:
          onMouseButtonDown(&wev0);
          break;
        case STATE_SELECTED:
          onSelectedMouseEvent(&wev0, true);
          break;
        }
      }
      break;
    case EV_INACT:
      break;
    case EV_DEVICE:
      oprc_dev(&(wev0.e), NULL, 0);
      break;
    case EV_MSG:
      clr_msg(MM_ALL, MM_ALL);
      break;
    case EV_MENU:
      if (wev0.s.pos.x >= SCREEN.hpixels - scrlbar_width){
        switch (state){
        case STATE_NORMAL:
          {
            W mi = menu.popupMenu(&wev0);
            if (mi > 0){
              selectMenu(&wev0, mi);
            }
          }
          break;
        case STATE_SELECTED:
          cancelSelection();
          break;
        }
      }
      else {
        switch (state){
        case STATE_NORMAL:
          onRightMouseButtonDown(&wev0);
          break;
        case STATE_SELECTED:
          onSelectedMouseEvent(&wev0, false);
          break;
        }
      }
      break;
    case EV_KEYDWN:
      if (state == STATE_SELECTED){
        cancelSelection();
      }
      onKeyDown(&wev0);
      break;
    }
  }
  
  display(false);
}

void MainWindow::openDocument(const LINK* link)
{
  bsfile = link;

  doc.clear();
  if (doc.load(link)) {
    errpanel(EPNL_LOADSKIP, 0);
  }

  mdenv.clear();
  doc.draw(mdenv.getGID(), cliprect.c.top, cliprect.c.bottom);
  draw();
}

void MainWindow::cancelSelection()
{
  doc.unselectAll();

  mdenv.clear();
  doc.draw(mdenv.getGID(), cliprect.c.top, cliprect.c.bottom);
  draw();

  state = STATE_NORMAL;
}

void MainWindow::onSelectedMouseEvent(WEVENT* ev, bool left)
{
  GID wgid, dgid;
  PNT ep;
  std::auto_ptr<SelectionStroke> stroke(sel_stroke);
  std::auto_ptr<ScribbleDoc::DrawingStrokePtrList> selected(sel_slist);

  wgid = wget_gid(wid);
  dgid = wsta_drg(wid, 0);
  gset_vis(wgid, cliprect);
  gset_vis(dgid, cliprect);
  if (!left){
    gcnv_rel(wgid, &(ev->s.pos));
  }

  if (wgid < 0 || dgid < 0){
    throw PaneledFailureException(EPNL_GID);
  }

  if (stroke->contain(ev->s.pos)){
    // 򤵤줿֥Ȥΰư(椫鳰)
    int np = stroke->getSize();
    SEL_RGN srgn;
    setrect(srgn.rgn.r, SHRT_MAX, stroke->getLowerBound(),
            SHRT_MIN, stroke->getUpperBound());
    srgn.sts = 0;

    draw();
    ScribbleDoc::DrawingStrokeConstPtrIterator iend = selected->end();
    for (ScribbleDoc::DrawingStrokePtrIterator iter = selected->begin();
         iter != iend; ++iter){
      (*iter)->drawSelection(dgid);
    }

    for (int i = 0; i < np; i++){
      int tx = stroke->getPoint(i).x;
      if (tx < srgn.rgn.r.c.left){
        srgn.rgn.r.c.left = tx;
      }
      if (tx > srgn.rgn.r.c.right){
        srgn.rgn.r.c.right = tx;
      }
    }
    
    int dh = 0, dv = 0;
    PNT sp = ev->s.pos;

    while ((left
            ? (wget_drg(&ep, ev) != EV_BUTUP)
            : ((ev->s.stat & ES_BUT2) != 0))){
      if (!left){
        wget_drg(&ep, ev);
      }
      if (ep.x == sp.x && ep.y == sp.y){
        continue;
      }
      dh += ep.x - sp.x;
      dv += ep.y - sp.y;
      draw();
      gmov_cor(dgid, sp.x - ep.x, sp.y - ep.y);
      for (ScribbleDoc::DrawingStrokePtrIterator iter = selected->begin();
           iter != iend; ++iter){
        (*iter)->drawSelection(dgid);
      }
    }
    wend_drg();

    doc.moveSelected(dh, dv);
    mdenv.clear();
    doc.draw(mdenv.getGID(), cliprect.c.top, cliprect.c.bottom);
    draw();
  }
  else {
    // 򤵤줿֥Ȥκ()
    bool del = false;
    SelectionStroke sel;

    sel.addPoint(ev->s.pos);

    while ((left
            ? (wget_drg(&ep, ev) != EV_BUTUP)
            : ((ev->s.stat & ES_BUT2) != 0))){
      if (!left){
        wget_drg(&ep, ev);
      }
      sel.addPoint(ep);
      sel.drawLatest(dgid);
      if (stroke->contain(ep)){
        del = true;
      }
    }
    wend_drg();

    if (del){
      doc.removeSelected();
      mdenv.clear();
      doc.draw(mdenv.getGID(), cliprect.c.top, cliprect.c.bottom);
      draw();
    }
    else {
      cancelSelection();
    }
  }

  state = STATE_NORMAL;
}
