/*!
 * @file MemoryDrawEnv.cc
 * @brief Ķ饹()
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2005-02-26, 
 *
 * Copyright (C) 2005, Yashiro Takeshi
 */

#include <basic.h>
#include <errcode.h>
#include <btron/dp.h>
#include <btron/libapp.h>
#include <bstdio.h>
#include <new>
#include "MemoryDrawEnv.h"
#include "Exception.h"
#include "databox.h"

extern DEV_SPEC SCREEN;

void MemoryDrawEnv::getCSPEC()
{
  bmp->planes = SCREEN.planes;
  bmp->pixbits = SCREEN.pixbits;
  if (GetStdCSPEC(&cspec, bmp->planes, bmp->pixbits) != ER_OK){
    throw PaneledFailureException(EPNL_GID);
  }
}

void MemoryDrawEnv::createBitmap()
{
  if (bmp->baseaddr[0] != 0){
    return; // already allocated
  }

  if (bmalloc(bmp) < 0){
    bmp->baseaddr[0] = 0;
    throw std::bad_alloc();
  }
}

void MemoryDrawEnv::releaseBitmap() throw()
{
  bfree(bmp);
  bmp->baseaddr[0] = 0;
}

GID MemoryDrawEnv::getGID() const
{
  if (gid < 0){
    gid = gopn_mem(NULL, bmp, (B*)&cspec);
    if (gid < 0){
      throw PaneledFailureException(EPNL_GID);
    }
  }
  return gid;
}

