/*!
 * @file ScribbleDoc.cc
 * @brief Scribbleɥ()
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2005-01-26, 
 *
 * Copyright (C) 2005, Yashiro Takeshi
 */

#include <basic.h>
#include <btron/file.h>
#include <bstdio.h>
#include <memory>
#include <vector>
#include <btron/dp.h>
#include <tad.h>
#include <btron/taddata.h>
#include "BufferedReader.h"
#include "BufferedWriter.h"
#include "ScribbleDoc.h"
#include "Stroke.h"
#include "Exception.h"

extern DEV_SPEC SCREEN;

W ScribbleDoc::draw(W gid, int lb, int ub) const {
  DrawingStrokeConstIterator i, iend = stroke.end();
  for (i = stroke.begin(); i != iend; ++i){
    i->draw(gid, lb, ub);
  }
}

bool ScribbleDoc::load(const LINK* src)
{
  bool ret = false, cont = true;
  W fd, type;

  // եΥץ
  fd = opn_fil(const_cast<LINK*>(src), F_READ, NULL);
  if (fd < 0){
    throw FileOpenException(fd, 0);
  }

  // TAD쥳ɤθ
  type = fnd_rec(fd, F_TOPEND, 1 << 1, 0, NULL);
  if (type < 0){
    cls_fil(fd);
    return false;
  }

  // ɤ߹
  std::auto_ptr<BufferedReader> fp(new BufferedReader(fd));
  
  while (!fp->isEOF()){
    B head[sizeof(LTADSEG)];
    TADSEG *th = reinterpret_cast<TADSEG*>(head);
    LTADSEG *lth = reinterpret_cast<LTADSEG*>(head);

    fp->read((B*)&(th->id), sizeof(th->id));

    if ((th->id & 0xff00) == 0xff00){
      fp->read((B*)&(th->len), sizeof(th->len));
      if (th->len != 0xffff){
        // ̾糧
        int id = th->id & 0xff;
        if (id == TS_FPRIM && cont) {
          std::vector<B> buf(th->len + 1);
          TF_SPLINE *pol = reinterpret_cast<TF_SPLINE*>(&(buf.at(0)));
          fp->read(&(buf.at(0)), th->len); 
          if (pol->subid == TF_SPLINE_SID){
            stroke.push_back(DrawingStroke(pol));
          }
          else {
            ret = true;
          }
          continue;
        }
        else if (id == TS_FIGEND){
          cont = false; // FIGENDθ褿TS_FPRIM̵
        }
        else if (id != TS_INFO && id != TS_FIG && id != TS_FDEF) {
          ret = true;
        }
        fp->skip(th->len);
      }
      else {
        // 顼
        fp->read((B*)&(lth->llen), sizeof(lth->llen));
        ret = true;
        fp->skip(lth->llen);
      }
    }
    else {
      // Ĺ
      ret = true;
    }
  }

  cls_fil(fd);
  return ret;
}

void ScribbleDoc::calcBounds(RECT *bound) const
{
  RECT rect = {{ 0, 0, SCREEN.hpixels, 0 }};

  DrawingStrokeConstIterator iend = stroke.end();
  for (DrawingStrokeConstIterator iter = stroke.begin();
       iter != iend; ++iter){
    int ub = iter->getUpperBound();
    int lb = iter->getLowerBound();
    if (ub > rect.c.bottom){
      rect.c.bottom = ub;
    }
    if (lb < rect.c.top){
      rect.c.top = lb;
    }
  }

  bound->c.top = rect.c.top;
  bound->c.left = rect.c.left;
  bound->c.bottom = rect.c.bottom;
  bound->c.right = rect.c.right;
}

void ScribbleDoc::save(const LINK* src) const 
{
  W fd, type, err;

  if (!modified){
    return;
  }
  
  // եΥץ
  fd = opn_fil(const_cast<LINK*>(src), F_WRITE, NULL);
  if (fd < 0){
    throw FileOpenException(fd, 0);
  }

  // TAD쥳ɤƬ˰ư(ʤϺ)
  type = fnd_rec(fd, F_TOPEND, 1 << 1, 0, NULL);
  if (type < 0){
    err = apd_rec(fd, NULL, 0, 1, 0, 0);
    if (err < 0){
      throw FileWriteException(fd, err);
    }
    err = see_rec(fd, -1, -1, NULL);
    if (err < 0){
      throw FileWriteException(fd, err);
    }
  }
  else {
    err = trc_rec(fd, 0);
    if (err < 0){
      throw FileWriteException(fd, err);
    }
  }

  std::auto_ptr<BufferedWriter> fp(new BufferedWriter(fd));

  RECT bound;
  calcBounds(&bound);

#if BIGENDIAN
#error Please arrange the little endian part to fit the big-endian.
#else
  static const TADSEG tsinfo_hd = { TS_INFO|0xff00, sizeof(H)*3 };
  static const INFOSEG tsinfo = { 0, 2, { 0x0121 } };
  TADSEG tsfig_hd = { TS_FIG|0xff00, sizeof(H)*10 + sizeof(W) };
  FIGSEG tsfig = { bound, bound, -120, -120, 0 };
  static const TADSEG tfpat_hd = { TS_FDEF|0xff00, TF_PATTERN_SIZE(1) };
  static const TF_PATTERN tfpat = { 0, TF_PATTERN_SID, 1, 1, 1, 1, { 0x10000000 } };
  static const COLOR tfpat_fgcol = 0x10000000;
  static const COLOR tfpat_bgcol = 0xffffffff; // transparent
  static const UH tfpat_mask = 7; // 100%
  TADSEG spl_hd = { TS_FPRIM|0xff00, TF_SPLINE_SIZE(1) };
  TF_SPLINE spl = { 0, TF_SPLINE_SID, (LN_SOLID<<8)|2, 1, 1, 0, 1 };
  static const TADSEG tsfigend_hd = { TS_FIGEND|0xff00, 0 };
#endif

  // TAD󥻥
  fp->write((B*)&tsinfo_hd, sizeof(H)*2);
  fp->write((B*)&tsinfo, tsinfo_hd.len);
  // ޷ϥ
  fp->write((B*)&tsfig_hd, sizeof(H)*2);
  fp->write((B*)&tsfig, tsfig_hd.len);
  // ѥ
  fp->write((B*)&tfpat_hd, sizeof(H)*2);
  fp->write((B*)&tfpat, sizeof(H)*5);
  fp->write((B*)&tfpat_fgcol, sizeof(COLOR));
  fp->write((B*)&tfpat_bgcol, sizeof(COLOR));
  fp->write((B*)&tfpat_mask, sizeof(UH));

  DrawingStrokeConstIterator iend = stroke.end();
  for (DrawingStrokeConstIterator iter = stroke.begin();
       iter != iend; ++iter){
    spl.np = iter->getSize();
    spl_hd.len = TF_SPLINE_SIZE(spl.np);
    // 
    fp->write((B*)&spl_hd, sizeof(H)*2);
    fp->write((B*)&spl, sizeof(H)*6);
    for (int i = 0; i < spl.np; i++){
      PNT p = iter->getPoint(i);
      fp->write((B*)&p, sizeof(PNT));
    }
  }  

  // ޷λ
  fp->write((B*)&tsfigend_hd, sizeof(UH)*2);
  fp->flush();
  cls_fil(fd);
  modified = false;

  return;
}

bool ScribbleDoc::select(DrawingStrokePtrList* selection, const SelectionStroke& selstroke, int lb, int ub)
{
  bool ret = false;

  DrawingStrokeConstIterator iend = stroke.end();
  for (DrawingStrokeIterator iter = stroke.begin();
       iter != iend; ++iter){
    if (iter->hasIntersection(lb, ub) && selstroke.contain(*iter)){
      ret = true;
      iter->select(true);
      selection->push_back(&(*iter));
    }
  }

  return ret;
}

void ScribbleDoc::unselectAll()
{
  DrawingStrokeConstIterator iend = stroke.end();
  for (DrawingStrokeIterator iter = stroke.begin();
       iter != iend; ++iter){
    iter->select(false);
  }
}

void ScribbleDoc::removeSelected()
{
  stroke.remove_if(std::mem_fun_ref(&DrawingStroke::isSelected));
  modified = true;
}

void ScribbleDoc::moveSelected(int dh, int dv)
{
  DrawingStrokeConstIterator iend = stroke.end();
  for (DrawingStrokeIterator iter = stroke.begin();
       iter != iend; ++iter){
    if (iter->isSelected()){
      iter->moveOffset(dh, dv);
      iter->select(false);
    }
  }
  modified = true;
}
