/*!
 * @file Stroke.cc
 * @brief ȥ()
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2005-01-26, 
 *
 * Copyright (C) 2005, Yashiro Takeshi
 */

#include <basic.h>
#include <btron/dp.h>
#include <vector>
#include <bstdio.h>
#include "Stroke.h"

//
// AbstractStroke Members
//

void AbstractStroke::addPoint(const PNT& p)
{
  if (p.y < y_lbound){
    y_lbound = p.y;
  }
  if (p.y > y_ubound){
    y_ubound = p.y;
  }
  if (getSize() > 0){
    int index = getSize()-1 + sz_header;
    int dx = poly[index].x - p.x;
    int dy = poly[index].y - p.y;
    if (dx*dx + dy*dy <= 1){
      return;
    }
  }
  poly.push_back(p);
  // manipulate the header field
  (reinterpret_cast<POLY*>(&(poly[0]))->size)++;
}

void AbstractStroke::moveOffset(int dh, int dv)
{
  if (y_ubound != INT_MIN){
    y_ubound += dv;
  }
  if (y_lbound != INT_MAX){
    y_lbound += dv;
  }

  int n = getSize();
  for (int i = 0; i < n; i++){
    poly[i + sz_header].x += dh;
    poly[i + sz_header].y += dv;
  }
}

//
// DrawingStroke Members
//

int DrawingStroke::lwidth = 3;
LATTR DrawingStroke::lattr = (LN_SOLID << 8) | lwidth;
PAT DrawingStroke::pat = {{0, 1, 1, 0x10000000, 0x10ffffff, FILL100}};
DCM DrawingStroke::dcm = G_STORE;
PAT DrawingStroke::pat_sel = {{0, 1, 1, 0x10ff0000, 0x10ffffff, FILL100}};

W DrawingStroke::draw(W gid, int lb, int ub) const
{
  if (!sel && hasIntersection(lb, ub)){
    return gdra_pln(gid, (POLY*)&(poly[0]), lattr, &pat, dcm);
  }
  return 0;
}

void DrawingStroke::clear()
{
  poly.clear();
  SelectionStroke();
}

//
// SelectionStroke Members
//

int SelectionStroke::lwidth = 2;
LATTR SelectionStroke::lattr = (LN_SOLID << 8) | lwidth;
PAT SelectionStroke::pat_line = {{0, 1, 1, 0x10ff0000, 0x10ffffff, FILL100}};
PAT SelectionStroke::pat_fill = {{0, 1, 1, 0x10ffc0c0, 0x10ffffff, FILL100}};
DCM SelectionStroke::dcm_line = G_STORE;
DCM SelectionStroke::dcm_fill = G_AND;

W SelectionStroke::draw(W gid, int lb, int ub) const
{
  if (isSelectionFinished()){
    // ¿ѷ
    W ret = gfil_pol(gid, (POLY*)&(poly[0]), &pat_fill, dcm_fill);
    if (ret < 0){
      return ret;
    }
    return gfra_pol(gid, (POLY*)&(poly[0]), lattr, &pat_line, dcm_line);
  }
  else {
    // 
    return gdra_pln(gid, (POLY*)&(poly[0]), lattr, &pat_line, dcm_line);
  }
}

W SelectionStroke::drawLatest(W gid) const
{
  int sz = getSize();
  if (!isSelectionFinished() && sz >= 2){
    return gdra_lin(gid, getPoint(sz-2), getPoint(sz-1),
                    lattr, &pat_line, dcm_line);
  }
  return 0;
}

void SelectionStroke::clear()
{
  poly.clear();
  SelectionStroke();
}

bool SelectionStroke::contain(const PNT& p) const
{
  const POLY* poly = getPoly();
  return (gpic_pol(p, const_cast<POLY*>(poly), 0) != 0);
}

bool SelectionStroke::contain(DrawingStroke& stroke) const
{
  const int np = 6;
  int index[np], prev_index = -1;

  int n = 0;
  int ss = stroke.getSize();

  //  np Ĥ(ʣƤindexϤȤȤФ)
  for (int i = 0; i < np; i++){
    int tmp = ss * (2 * i + 1) / (np * 2);
    if (prev_index != tmp){
      prev_index = tmp;
      index[n] = tmp;
      n++;
    }
  }

  // äƤ뤫Τ
  for (int i = 0; i < n; i++){
    if (!contain(stroke.getPoint(index[i]))){
      return false;
    }
  }
  return true;
}

